//
//  Spider+Promises.swift
//  Pods
//
//  Created by Mitch Treece on 12/13/16.
//  Copyright (c) 2017 Mitch Treece. All rights reserved.
//

import PromiseKit

extension Spider {
    
    /**
     Performs a `SpiderRequest`.
     - Parameter request: The `SpiderRequest` to perform.
     - Returns: A promise over `SpiderResponse`.
     */
    public func perform(_ request: SpiderRequest) -> Promise<SpiderResponse> {
        
        return Promise<SpiderResponse> { (fulfill, reject) in
            perform(request) { (response) in
                fulfill(response)
            }
        }
        
    }
    
    /**
     Performs a GET request with various configuration options.
     - Parameter path: The endpoint path to append to the global base URL **or** a fully qualified URL (if no global base URL is specified).
     ```
     "/users/12345"
     "http://base.url/v1/users/12345"
     ```
     - Parameter parameters: An optional param object to be passed along with the request.
     - Parameter auth: An optional authorization type to use for this request. This will _override_ Spider's global authorization type. If no authorization type is provided, the request will fallback to Spider's global authorization type.
     - Returns: A promise over `SpiderResponse`.
     */
    public func get(_ path: String, parameters: Any? = nil, auth: SpiderAuth? = nil) -> Promise<SpiderResponse> {
        
        let request = SpiderRequest(method: .get, path: path, parameters: parameters, auth: auth)
        return Promise<SpiderResponse> { (fulfill, reject) in
            _ = perform(request).then { (response) -> Void in
                fulfill(response)
            }
        }
        
    }
    
    /**
     Performs a POST request with various configuration options.
     - Parameter path: The endpoint path to append to the global base URL **or** a fully qualified URL (if no global base URL is specified).
     ```
     "/users/12345"
     "http://base.url/v1/users/12345"
     ```
     - Parameter parameters: An optional param object to be passed along with the request.
     - Parameter auth: An optional authorization type to use for this request. This will _override_ Spider's global authorization type. If no authorization type is provided, the request will fallback to Spider's global authorization type.
     - Returns: A promise over `SpiderResponse`.
     */
    public func post(_ path: String, parameters: Any? = nil, auth: SpiderAuth? = nil) -> Promise<SpiderResponse> {
        
        let request = SpiderRequest(method: .post, path: path, parameters: parameters, auth: auth)
        return Promise<SpiderResponse> { (fulfill, reject) in
            _ = perform(request).then { (response) -> Void in
                fulfill(response)
            }
        }
        
    }
    
    /**
     Performs a PUT request with various configuration options.
     - Parameter path: The endpoint path to append to the global base URL **or** a fully qualified URL (if no global base URL is specified).
     ```
     "/users/12345"
     "http://base.url/v1/users/12345"
     ```
     - Parameter parameters: An optional param object to be passed along with the request.
     - Parameter auth: An optional authorization type to use for this request. This will _override_ Spider's global authorization type. If no authorization type is provided, the request will fallback to Spider's global authorization type.
     - Returns: A promise over `SpiderResponse`.
     */
    public func put(_ path: String, parameters: Any? = nil, auth: SpiderAuth? = nil) -> Promise<SpiderResponse> {
        
        let request = SpiderRequest(method: .put, path: path, parameters: parameters, auth: auth)
        return Promise<SpiderResponse> { (fulfill, reject) in
            _ = perform(request).then { (response) -> Void in
                fulfill(response)
            }
        }
        
    }
    
    /**
     Performs a PATCH request with various configuration options.
     - Parameter path: The endpoint path to append to the global base URL **or** a fully qualified URL (if no global base URL is specified).
     ```
     "/users/12345"
     "http://base.url/v1/users/12345"
     ```
     - Parameter parameters: An optional param object to be passed along with the request.
     - Parameter auth: An optional authorization type to use for this request. This will _override_ Spider's global authorization type. If no authorization type is provided, the request will fallback to Spider's global authorization type.
     - Returns: A promise over `SpiderResponse`.
     */
    public func patch(_ path: String, parameters: Any? = nil, auth: SpiderAuth? = nil) -> Promise<SpiderResponse> {
        
        let request = SpiderRequest(method: .patch, path: path, parameters: parameters, auth: auth)
        return Promise<SpiderResponse> { (fulfill, reject) in
            _ = perform(request).then { (response) -> Void in
                fulfill(response)
            }
        }
        
    }
    
    /**
     Performs a DELETE request with various configuration options.
     - Parameter path: The endpoint path to append to the global base URL **or** a fully qualified URL (if no global base URL is specified).
     ```
     "/users/12345"
     "http://base.url/v1/users/12345"
     ```
     - Parameter parameters: An optional param object to be passed along with the request.
     - Parameter auth: An optional authorization type to use for this request. This will _override_ Spider's global authorization type. If no authorization type is provided, the request will fallback to Spider's global authorization type.
     - Returns: A promise over `SpiderResponse`.
     */
    public func delete(_ path: String, parameters: Any? = nil, auth: SpiderAuth? = nil) -> Promise<SpiderResponse> {
        
        let request = SpiderRequest(method: .delete, path: path, parameters: parameters, auth: auth)
        return Promise<SpiderResponse> { (fulfill, reject) in
            _ = perform(request).then { (response) -> Void in
                fulfill(response)
            }
        }
        
    }
    
}
